import os
import subprocess
import hashlib
import shutil
import filecmp
import re

from .xmltotext import Transform


class TransformToPDF():
    
    def __init__(self, inname, outname=None, force=False, quiet=False):
        self.inname = inname
        self.basename = self.inname[:self.inname.find('.')]
        self.working_directory = os.path.join(os.curdir, 'working_directory')
        self.outtex = os.path.join(self.working_directory, self.basename + '.tex')
        if not outname:
            self.outname = self.basename + '.pdf'
        else:
            self.outname = outname
        self.force = force
        self.quiet = quiet
        self.resourcesdir = os.path.join(os.curdir, 'resources')

    def transform(self):
        self.check_requirements()
        latex = self.make_latex()
        pdf_path = self.make_pdf(latex)
        self.on_pdf_creation(pdf_path)

    def check_requirements(self):
        if not os.path.exists(self.working_directory):
            os.mkdir(self.working_directory)

        resource_files = ['references.bib',
                      'personlist.xml',
                      'introduction.tex',
                      'latex_preamble.tex']
        for name in resource_files:
            inp = os.path.join(self.resourcesdir, name)
            outp = os.path.join(self.working_directory, name)
            assert os.path.isfile(inp)

            no_out = not os.path.isfile(outp)
            if no_out:
                self._copy_required_file(inp, outp)
            elif not filecmp.cmp(inp, outp):
                self._copy_required_file(inp, outp)

        i_s_p = os.path.join(self.working_directory, self.basename + '.mst')
        if not os.path.isfile(i_s_p):
            with open(i_s_p, 'w') as isty:
                isty.write(self.index_style())


    def _copy_required_file(self, in_, out_):
        shutil.copy2(in_, out_)
        self.force = True
        if os.path.basename(in_) == 'personlist.xml':
            picklepath = os.path.join(self.working_directory,
                                'personlist.pickle')
            if os.path.isfile(picklepath):
                os.unlink(picklepath)


    def make_latex(self):
        t = self._get_transformer()
        text = t()
        pre = self.latex_preamble()
        front = self.latex_front_matter()
        after = self.latex_end_matter()

        header = pre.replace(r'\begin{document}', front)
        _all = '\n'.join([header, text, after])

        return self._after_creation(_all)

    def make_pdf(self, tex):
        if not self.check_run(tex):
            with open(self.outtex, 'w') as o:
                o.write(tex)
            result = self.call_latex()
        else:
            print('Has not changed since the last run')
        return os.path.join(self.working_directory, self.basename + '.pdf')        

    def on_pdf_creation(self, pdfpath):
        shutil.copy(pdfpath, self.outname)

    def _get_transformer(self):
        transformer = Transform(self.inname, self.working_directory)
        return transformer.transform

    def check_run(self, tex):
        if not self.force:
            if os.path.exists(self.outtex):
                with open(self.outtex, 'rb') as exists:
                    exists = exists.read()
                existing_hash = hashlib.md5(exists).digest()
                new_hash = hashlib.md5(tex.encode('utf8')).digest()
                return existing_hash == new_hash
        return False


    def latex_front_matter(self):
        return '\n'.join([
            '\\begin{document}',
            '\\frontmatter',
            '\\tableofcontents',
            '\setcounter{secnumdepth}{-2}',
            #'\include{./introduction}',
            '\mainmatter',
            '\lineation{page}',
            '\setlength{\stanzaindentbase}{30pt}',
            '\setstanzaindents{3,1,1}',
            '\setcounter{stanzaindentsrepetition}{1}',
            '\\newcommand*{\startstanzahook}{\\vspace{9pt}}',
            '\def\endstanzaextra{\\vspace{9pt}}',
            #'\\allsectionsfont{\\normalsize}',
            '\\beginnumbering',
        ])


    def latex_preamble(self):
        prefile = os.path.join(self.working_directory, 'latex_preamble.tex')
        with open(prefile) as r:
            return r.read()

    def latex_end_matter(self):
        return '\n'.join([
            '\endnumbering',
            '\\backmatter'
            '\clearpage',
            #'\\rfoot{\\textsc{References} / \\thepage}',
            '\printbibliography',
            '\clearpage',
            #'\\rfoot{\\textsc{Editorial Practice} / \\thepage}',
            #'\\rfoot{\\textsc{Index / \\thepage}}',
            '\printindex',
            '\end{document}',
        ])


    def _after_creation(self, text):
        for k, v in {
            'i.e. ': 'i.e.\ ',
            'e.g. ': 'e.g.\ ',
            ' v. ': ' v.\ ',
            ' w. ': ' w.\ ',
            ' wh. ': ' wh.\ ',
            'MG. ': 'MG.\@ ',
            'HQ. ': 'HQ.\@ ',
            'YMCA. ': 'YMCA.\@ ',
            'ADS. ': 'ADS.\@ ',
            'RMT. ': 'RMT.\@ ',
            'NZ. ': 'NZ.\@ ',
            'M.T. ': 'M.T.\@ ',
            'Horowhenua': 'Horo\-whenua',
            'HQ. C.O.': 'HQ.\@ C.O.',
            'C.O. Battalion dinner': 'C.O.\@ Battalion dinner',
            'GA. So': 'GA.\@ So',
            'Trémouille': 'Tré\-mouille',
            'or M&V.': 'or M\\&V.',
            '∴': '\\texttherefore{}', 
        }.items():
            text = text.replace(k, v)
        for char in '.,!)':
            text = text.replace('-' + char, '---' + char)
        hyphensubs = [(r'(?<=\d)-(?=\d)', '--'), # hyphens between numbers to en-dashes
                    (r'(?<=\s)-(?=\s)', '---'), # hyphens surrounded by whitespace to em-dashes.
                    ]
        for sub in hyphensubs:
            text = re.sub(sub[0], sub[1], text)
        # Hyphens in our citations!
        text = re.sub(r'\\pageref\{(.*?)--(.*?)\}',
            r'\\pageref{\1-\2}', text)
        text = re.sub(r'\\autocite\{(.*?)--(.*?)\}',
            r'\\autocite{\1-\2}', text)
        text = re.sub(r'\\label\{(.*?)--(.*?)\}',
            r'\\label{\1-\2}', text)
        text = text.replace('----)', '--)')
        text = re.sub(r'\ +', ' ', text) # Be a tidy kiwi.
        text = re.sub(r'\n\ +', '\n', text)
        text = re.sub(r'\n\n+', '\n\n', text)
        return text


    def index_style(self):
        return '\n'.join(['headings_flag 1',
            'heading_prefix "{\\\\bfseries "',
            'heading_suffix "}\\\\nopagebreak\\n"',
            ])

    def call_latex(self):
        options = ['-bibtex', # run biber for references
           '-cd', # change to working_directory to run
           '-f', # force through errors
           '-g', # run even if unchanged.
           '-pdf',]
        latexmk_command = ['latexmk'] + options + [self.outtex]
        if self.quiet:
            with open(os.devnull, "w") as fnull:
                return subprocess.call(latexmk_command, stdout = fnull, stderr = fnull)
        return subprocess.call(latexmk_command)


import argparse
import os
import shutil
from tei_transformer import transform

def get_args():
    parser = argparse.ArgumentParser(
        description="Transform a TEI-encoded critical edition into a pdf.")
    parser.add_argument("-t", "--transform",
                        help="TEI file to transform")
    parser.add_argument("-o", "--outputname",
                        help="Filename of the transformed file.",
                        default=None)
    parser.add_argument("-f", "--force",
                        help="Force recompilation of a latex file,\
                         even if unchanged.",
                        action="store_true")
    parser.add_argument('-q', '--quiet',
                        help="Run quietly",
                        action="store_true")
    return parser.parse_args()


def main():
    args = get_args()
    if args.transform:
        transformer = transform.TransformToPDF(args.transform, args.outputname, args.force, args.quiet)
        transformer.transform()

if __name__ == '__main__':
    main()



import os
import sys
import pickle
from lxml import etree

import collections

from .tags import TEITag

class Transform():

    def __init__(self, filepath, working_directory, use_persdict=True):

        self.filepath = filepath
        self.working_directory = working_directory

        self.parser = Parser()
        if use_persdict:
            self.persdict = PersDict(self.working_directory, self.parser)
        else:
            self.persdict = None

    def transform(self):
        tree = self.parser(self.filepath)
        root = tree.getroot()
        body = root.find('.//{*}body') # {*} is to map any namespace
        assert body is not None
        tree = self.parser.transform_tree(body, self.persdict)
        return '\n'.join(tree.itertext()).strip()

class PersDict(collections.UserDict):

    def __init__(self, working_directory, parser):
        self.picklepath = os.path.join(working_directory, 'personlist.pickle')
        self.path = os.path.join(working_directory, 'personlist.xml')
        self.parser = parser
        self._get_persdict()

    def _get_persdict(self):
        nelson = self._from_pickle()
        if nelson:
            super().__init__(nelson)
        else:
            torpedoes = self._make_persdict()
            super().__init__(torpedoes)
            self._to_pickle()

    def _make_persdict(self):
        people = self.parser(self.path).getroot().iter('{*}person')
        return self._process_people(people)

    def _process_people(self, people):
        persdict = dict()
        for person in people:
            xml_id, person_dict = self._first_run(person)
            persdict[xml_id] = person_dict
        return self._second_run(persdict)

    # PROCESSING TAGS

    def _first_run(self, tag):
        person = dict()
        xml_id = self._xml_id(tag)
        person['indexname'] = self._indexname(tag)
        person['description'] = tag
        return xml_id, person

    def _second_run(self, persdict):
        for xml_id, person in persdict.items():
            tag = person['description']
            person['indexonly'] = self._indexonly(tag)
            person['description'] = self._get_description(tag, persdict)
        return persdict


    # INFO FROM TAGS

    def _xml_id(self, tag):
        return tag.attrib.get('%sid' % '{http://www.w3.org/XML/1998/namespace}')

    def _indexname(self, tag):
        firstnames, surnames = self.__get_names(tag)
        return ', '.join([' '.join(reversed(surnames)), firstnames])

    def _indexonly(self, tag):
        return tag.find('{*}indexonly') is not None # Non-TEI addition

    def _get_description(self, tag, persdict):
        name = self._descriptionname(tag)
        dates = self._dates(tag)
        biography = self._biography(tag, persdict)
        return ' '.join([name, dates, biography]).strip()

    # Getting description...

    def _descriptionname(self, tag):
        firstnames, surnames = self.__get_names(tag)
        surnames = ' '.join(surnames)
        return ' '.join([firstnames, surnames])

    def _dates(self, tag):
        birth = self.__find_string(tag, 'birth')
        death = self.__find_string(tag, 'death')
        return '(%s--%s)' % (birth[0], death[0])

    def _biography(self, tag, persdict):
        trait = tag.find('{*}trait')
        if trait is None:
            return ''
        traitpara = trait.find('{*}p')
        if traitpara is not None:
            if traitpara.getchildren() or traitpara.text:
                self.parser.transform_tree(traitpara, persdict, in_body=False)
                return traitpara.text.strip()
        return ''

    def __find_string(self, parent, searchterm):
        searchstring = '{*}' + searchterm
        matches = parent.findall(searchstring)
        return [str((name.text or '').strip()) for name in matches]

    def __get_names(self, tag):
        persname = tag.find('{*}persName')
        forenames = self.__find_string(persname, 'forename')
        addnames = ["`%s'" % x for x in self.__find_string(persname, 'addName')]
        surnames = self.__find_string(persname, 'surname')
        firstnames = ' '.join(forenames + addnames)
        return firstnames, surnames


    # PICKLING

    def _from_pickle(self):
        try:
            with open(self.picklepath, 'rb') as p:
                return pickle.load(p)
        except (EOFError, FileNotFoundError) as e:
            if e == EOFError: # i.e. something went wrong with pickling
                os.unlink(self.picklepath)

    def _to_pickle(self):
        with open(self.picklepath, 'wb') as p:
            pickle.dump(self.data, p)



class Parser():

    def __init__(self):
        lookup = etree.ElementNamespaceClassLookup()
        self.parser = etree.XMLParser(ns_clean=True, remove_comments=True, remove_blank_text=True)
        self.parser.set_element_class_lookup(lookup)
        namespace = lookup.get_namespace('http://www.tei-c.org/ns/1.0')
        # We set TEITag as the default class for the parser.
        namespace[None] = TEITag
        # Map the target to handling class in the lxml parser.
        for cls in self.tag_classes():
            namespace[cls.target] = cls

    def __call__(self, textpath):
        try:
            tree = etree.parse(textpath, self.parser)
        except FileNotFoundError:
            print('Could not find %s.' % textpath)
            sys.exit(1)
        except etree.XMLSyntaxError as err:
            print('An error occurred parsing {textpath}:\n\'{err}\''.format(textpath=textpath, err=err))
            sys.exit(1)
        return tree

    def _tag_classes(self, cls=TEITag):
        for subcls in cls.__subclasses__():
            yield from self._tag_classes(cls=subcls)
            if hasattr(subcls, 'target') and subcls.target:
                yield subcls


    def tag_classes(self):
        """You may well want to override this to include your own..."""
        return self._tag_classes()


    def transform_tree(self, tree, persdict, in_body=True):
        taglist = list(tree.getiterator('*'))
        taglist.sort(key=lambda tag: tag.descendants_count())
        for tag in taglist:
            if tag.localname == 'persName':
                if not in_body:
                    tag.process(persdict, in_body=False)
                else:
                    tag.process(persdict)
            else:
                tag.process()
        return tree 




from lxml import etree
import calendar

LANGUAGES = {
    'it': 'italian',
    'de': 'german',
    'la': 'latin',
    'fr': 'french',
    'gr': 'greek',
}


class ImplementationError(Exception):
    pass

class TEITag(etree.ElementBase):

    def _init(self):
        self._localname = None
        if self.text:
            self.text = self._text_to_latex(self.text)
        if self.tail:
            self.tail = self._text_to_latex(self.tail)

    def process(self, persdict=None, in_body=True):

        # Check to see we won't accidentally wipe something out.
        not_good_children = self.localname not in ['choice', 'app']
        if not self.no_children() and not_good_children:
            self._raise()
        # Now on with the replacement.
        if self.localname == 'persName':
            replacement = self.get_replacement(persdict, in_body)
        else:
            replacement = self.get_replacement()
        if not replacement:
            return # i.e. don't touch this; it will be processed later by a parent.
        else:
            self.replace_w_str(replacement)


    @property
    def localname(self):
        if not self._localname:
            self._localname = etree.QName(self).localname
        return self._localname

    def no_children(self):
        return len(self) == 0
    
    def _text_to_latex(self, text):
        return text

    def __str__(self):
        return etree.tounicode(self, with_tail=False)

    def descendants_count(self):
        if self.no_children():
            return 0
        else:
            return sum(1 for _ in self.iterdescendants())

    def delete(self):
        parent = self.add_to_previous(self.tail)
        parent.remove(self)

    def get_replacement(self):
        raise NotImplementedError(self)

    def unwrap(self):
        children = list(self.iterchildren(reversed=True))
        if not len(children):
            self.replace_w_str(self.text)
        else:
            parent = self.getparent()
            index = parent.index(self)
            last_child = children[-1]
            last_child.tail = self.textjoin(last_child.tail, self.tail)
            parent = self.add_to_previous(self.textjoin(self.text, self.tail))
            for child in children:
                parent.insert(index, child)

    def replace_w_str(self, replacement):
        replacement = self.textjoin(replacement, self.tail)
        parent = self.add_to_previous(replacement)
        parent.remove(self)

    def add_to_previous(self, addition):
        previous = self.getprevious()
        parent = self.getparent()
        if previous is not None:
            previous.tail = self.textjoin(previous.tail, addition)
        else:
            parent.text = self.textjoin(parent.text, addition)
        return parent

    @staticmethod
    def textjoin(a, b):
        return ''.join([(a or ''), (b or '')])

    def _raise(self):
        raise ImplementationError(self)



    def required_key(self, key):
        x = self.get(key)
        if x:
            return x
        self._raise()

    def ex_ref(self, key):
        key = self.required_key(key)
        return key[1:]

    def rend(self, required=False):
        if required:
            return self.required_key('rend')
        return self.get('rend')

    def n(self, required=False):
        if required:
            return self.required_key('n')
        return self.get('n')
    
    def key_or_empty(self, key):
        return self.get(key) or ''


# STANDARD

class Head(TEITag):

    target = 'head'

    def get_replacement(self):
    
        """ Return the replacement for a tag of type <head>"""

        # Make explicit the desire not to indent following paragraph.
        if not self.tail or not self.tail.strip():
            next_sibling = self.getnext()
            if next_sibling is not None and next_sibling.localname == 'p':
                if next_sibling.get('rend') == None:
                    next_sibling.text = '\\noindent %s' % (next_sibling.text or '')

        parent_attrs = self.getparent().attrib
        div_type = parent_attrs['type']
        # if div_type == None: # Part
        #     return self.process_head_level_one(contents)
        if div_type == 'title': # Chapter
            return self.process_head_level_two()

        try:
            identifier = parent_attrs['%sid' % '{http://www.w3.org/XML/1998/namespace}']
        except KeyError:
            self._raise()

        if div_type == 'diaryentry': # Section
            return self.process_head_level_three(identifier)
        elif div_type == 'diaryentrysection': # Subsection
            return self.process_head_level_four(identifier)
        self._raise()

    def process_head_level_one(self):
        return '\\part{%s}' % self.text

    def process_head_level_two(self):
        return '\\pstart\n\\eledchapter*{%s}\n\\pend' % self.text

    def process_head_level_three(self, identifier):
        if identifier is None:
            self._raise()
        month = identifier[:3]
        # Assumes gregorian calendar
        for index, string in enumerate(calendar.month_abbr):
            if string == month:
                month = calendar.month_name[index]
                break
        date = identifier[3:identifier.find('_')]
        year = identifier[-4:]        
        return '\n\\pstart\n\
               \\eledsection[%s %s %s]{%s}\n\
               \\label{%s}\n\
               \\pend'\
               % (date, month, year, self.text, identifier)

    def process_head_level_four(self, identifier):
        return '\\pstart\n\\eledsubsection{%s}\n\\label{%s}\n\\pend'\
                % (self.text, identifier)

    def process_head_level_five(self, identifier):
        return '\\pstart\n\\eledsubsubsection{%s}\n\\label{%s}\n\\pend'\
                % (self.text, identifier)


class TextualNote(TEITag):
    
    target = 'note'

    def get_replacement(self):
        if self.empty():
            return

        marker = self.get_marker()
        assert marker is not False
        assert self.get('type') == 'annotation'
        marker.replace_w_str('\\annotation{')
        return '}{%s}' % self.text

    def empty(self):
        return self.text is None and len(self.getchildren()) == 0

    def get_marker(self):
        skip = 0
        for n in self._previous_notes():
            if n.text:
                skip += 1
            else:
                if skip == 0:
                    return n
                skip = skip - 1
        return False

    def _previous_notes(self):
        for sib in self.itersiblings(preceding=True):
            if sib.localname == 'note':
                yield sib


class Div(TEITag):
    target = 'div'

    def get_replacement(self):
        divtype = self.get('type')
        if not divtype:
            return self.unwrap()
        elif divtype in ['month', 'year']:
            date = self.n(required=True)
            if divtype == 'month':
                year = self._find_year()
                assert year
                before = '\n\\addcontentsline{toc}{chapter}\
                      {%s %s}' % (date, year)
            elif divtype == 'year':
                before = '\n\\addcontentsline{toc}{part}{%s}' % date

            self.add_to_previous(before)
        return self.unwrap()

    def _find_year(self):
        for parent in self.iterancestors():
            if parent.get('type') == 'year':
                return parent.n(required=True)


class Paragraph(TEITag):
    target = 'p'

    def get_replacement(self):
        # if not self.text:
        #     return self.delete()
        try:
            paratext = self.text.strip()
        except AttributeError:
            self._raise()
        paratext = paratext.replace('\n', ' ')
        rend = self.rend()
        if rend == 'noindent':
            indent = '\\noindent '
        elif rend == 'indent':
            indent = '\\indent '
        elif rend == 'doubleindent':
            indent = '\\doubleindent '
        else:
            # Our headers are wrapped in an eledmac paragraph.
            # So latex doesn't know not to indent them.
            previous = self.getprevious()
            if previous is not None and previous.localname == 'head':
                indent = '\\noindent '
            else:
                indent = ' '

        return ''.join(['\n\\pstart', indent, paratext, '\\pend'])


class VerseLineGroup(TEITag):
    target = 'lg'

    def get_replacement(self):
        if self.tail:
            self.tail = self.tail.strip()
        stanzatext = self.text.strip()
        stanzatext = stanzatext.rstrip('&')
        return '\n\n\\stanza\n%s \\&\n\n' % stanzatext


class VerseLine(TEITag):
    target = 'l'

    def get_replacement(self):
        if self.tail:
            self.tail = self.tail.strip()
        if self.text:
            self.text = self.text.strip()
        return '%s &\n' % self.text

class Foreign(TEITag):
    target = 'foreign'

    def get_replacement(self):
        langcode = self.get('%slang' % '{http://www.w3.org/XML/1998/namespace}')
        language = LANGUAGES.get(langcode)
        if language:
            return '\\text%s{%s}' % (language, self.text)
        return self.text

class FloatingText(TEITag):
    target = 'floatingText'

    def get_replacement(self):
        floattype = self.get('type')
        if floattype == 'verse':
            return '\n\\vspace{5mm}%s\\vspace{5mm}\n' % self.text
        elif floattype == 'addition':
            return '\\pstart %s\\pend' % self.text
        self._raise()

class PersName(TEITag):
    target = 'persName'

    def _init(self):
        super()._init()
        self.in_body_text = True
        for x in self.iterancestors('{*}note'):
            self.in_body_text = False
            break

    def get_replacement(self, persdict, in_body):
        ref = self.ex_ref('ref')
        if not ref:
            self._raise()
        elif ref == '??':
            return '\\unknownperson{%s}' % self.text

        if not in_body:
            self.in_body_text = False

        person = persdict[ref]
        indexonly = person.get('indexing_only')
        indexname = person['indexname']
        if indexonly or not self.in_body_text:
            if not self.in_body_text:
                indexname += '|innote'
            return '\\indexperson{%s}{%s}' % (indexname, self.text)
        else:
            description = person.get('description')
            return '\\person{%s}{%s}{%s}{%s}' % (ref, indexname, description, self.text)



class Space(TEITag):
    target = 'space'

    def get_replacement(self):
        n = self.n()
        if n == 'vertical':
            return '\\bigskip'
        elif n == 'horizontal':
            return '\\hfill{}'
        else:
            return '\\qquad{}'
        self._raise()


class PageBreak(TEITag):
    target = 'pb'

    def get_replacement(self):

        pagenumber = self.attrib['n']
        pagenumber = pagenumber.lstrip('0')
        if pagenumber == '1':
            return self.delete()

        if self.tail is not None:
            self.tail = self.tail.lstrip()
        previous = self.getprevious()
        if previous is not None:
            if previous.tail:
               previous.tail = previous.tail.rstrip()

        
        parent = self.getparent()
        if parent.localname in ['p', 'lg'] :
            return ' \\intextpagebreak{[%s]} ' % pagenumber
        else:
            return '\n\\floatpagebreak{[%s]}\n' % pagenumber


class Label(TEITag):
    target = 'label'

    def get_replacement(self):
        n = self.n(required=True)
        return '\\label{%s}' % n


class Deletion(TEITag):
    target = 'del'

    def get_replacement(self):
        if self.get('resp'):
            return self.delete()
        elif self.get('hand'):
            return '\\sout{%s}' % self.text
        else:
            self._raise()

class Add(TEITag):
    target = 'add'

    def get_replacement(self):
        return '\\addition{%s}' % self.text

class FilterTag(TEITag):

    def filterchildren(self, targets):
        for x in self.iterchildren():
            if x.localname in targets:
                yield x.localname, x
                targets.remove(x.localname)

        if len(targets) != 0:
            self._raise()


class Choice(FilterTag):
    target = 'choice'

    def get_replacement(self):
        for name, tag in self.filterchildren(['sic', 'corr']):
            if name == 'sic':
                sic_text = tag.text
            elif name == 'corr':
                corr_text = tag.text

        return '\\correction{%s}{%s}' % (corr_text, sic_text)

class TextualVariant(FilterTag):
    target = 'app'

    def get_replacement(self):
        rdgs = []
        for name, tag in self.filterchildren(['lem', 'rdg']):
            if name == 'lem':
                lem_text = tag.text
            elif name == 'rdg':
                rdgs.append(tag)
        
        rdgs = map(self.rdgs_map, rdgs)
        return '\\variants{%s}{%s}' % (lem_text, ' '.join(rdgs))

    @staticmethod
    def rdgs_map(rdg):
        try:
            wit = rdg.attrib['wit']
        except KeyError:
            rdg._raise()
        return '\\wit{%s}{%s}' % (rdg.text, wit)



# FURTHER SUBCLASSING.

class FmtTag(TEITag):

    def _wrap(self, before, after, text=None):
        if text is None:
            text = self.text
        return '%s%s%s' % (before, text, after)

    def square_brackets(self, text=None):
        return self._wrap('[', ']', text)

    def single_quotes(self, text=None):
        return self._wrap('`', "'", text)

    def double_quotes(self, text=None):
        return self._wrap("``", "''", text)

    def emph(self, text=None):
        if text is None:
            text = self.text
        return '\\emph{%s}' % text

    def superscript(self, text=None):
        if text is None:
            text = self.text
        return '\\textsuperscript{%s}' % text

    def smallcaps(self, text=None):
        if text is None:
            text = self.text
        return '\\textsc{%s}' % text

    def handle_none(self):
        return self.text

    def _singles(self):
        return ['single']

    def _doubles(self):
        return ['double']

    def _emphs(self):
        return ['emph', 'italic', 'underscore', 'underline']

    def _supers(self):
        return ['super', 'upper']

    def _smcps(self):
        return ['smcp']

    def _none(self):
        return [None, 'none', 'None']

class SoCalled(FmtTag):
    target = 'soCalled'

    def get_replacement(self):
        return self.single_quotes()

class Ptr(FmtTag):
    
    target = 'ptr'

    def get_replacement(self):
        target = self.ex_ref('target')
        kind = self.get('type')

        bib = kind == 'bibliog'
        crossref = kind == 'crossref'

        if bib:
            pre = self.key_or_empty('pre')
            if pre:
                pre = self.square_brackets(pre)
            n = self.key_or_empty('n')
            if pre or n:
                n = self.square_brackets(n)
            return ' \\autocite%s%s{%s}' % (pre, n, target)
        elif crossref:
            return '\\pageref{%s}' % target

        self._raise()


class Supplied(FmtTag):
    target = 'supplied'

    def get_replacement(self):
        return self._wrap('«', '»')


class RendTag(FmtTag):

    def fmt_lists(self):
        return [(self._emphs(), self.emph),
                (self._singles(), self.single_quotes),
                (self._doubles(), self.double_quotes),
                (self._supers(), self.superscript),
                (self._smcps(), self.smallcaps),
                (self._none(), self.handle_none)]


    def _rendery(self, *args, **kwargs):
        rend = self.rend(*args, **kwargs)
        for targets, func in self.fmt_lists():
            if rend in targets:
                return func()
        self._raise()


class Bibl(RendTag):
    target = 'bibl'

    def get_replacement(self):
        return self._rendery()

class Hi(RendTag):
    target = 'hi'

    def get_replacement(self):
        return self._rendery(required=True)

class Quote(RendTag):
    target = 'q'

    def handle_none(self):
        return self.single_quotes()

    def get_replacement(self):
        return self._rendery()

# SIMPLE ONES:

class DeleteMe(TEITag):

    def get_replacement(self):
        return self.delete()

class DontTouchMe(TEITag):

    def get_replacement(self):
        return None

class ReplaceMeWText(TEITag):

    def get_replacement(self):
        return self.text

class UnWrapMe(TEITag):

    def get_replacement(self):
        return self.unwrap()

class Body(ReplaceMeWText):
    target = 'body'

class Corr(DontTouchMe):
    target = 'corr'

class Sic(DontTouchMe):
    target = 'sic'

class Lemma(DontTouchMe):
    target = 'lem'

class VariantReading(DontTouchMe):
    target = 'rdg'

class Time(ReplaceMeWText):
    target = 'time'

class LineBreak(DeleteMe):
    target = 'lb'

class Subst(UnWrapMe):
    target = 'subst'

class List(ReplaceMeWText):
    target = 'list'

class ListItem(ReplaceMeWText):
    target = 'item'

class Name(ReplaceMeWText):
    target = 'name'
